# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.op.op import Op
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class TensorOp(Op):
    """An Operator class which takes and returns tensor data.

    These Operators are used in fe.Network to perform graph-based operations like neural network training.
    """
    def forward(self, data: Union[Tensor, List[Tensor]], state: Dict[str, Any]) -> Union[Tensor, List[Tensor]]:
        """A method which will be invoked in order to transform data.

        This method will be invoked on batches of data.

        Args:
            data: The batch from the data dictionary corresponding to whatever keys this Op declares as its `inputs`.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The `data` after applying whatever transform this Op is responsible for. It will be written into the data
            dictionary based on whatever keys this Op declares as its `outputs`.
        """
        return data

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        """A method which will be invoked during Network instantiation.

        This method can be used to augment the natural __init__ method of the TensorOp once the desired backend
        framework is known.

        Args:
            framework: Which framework this Op will be executing in. One of 'tf' or 'torch'.
            device: Which device this Op will execute on. Usually 'cuda:0' or 'cpu'. Only populated when the `framework`
                is 'torch'.
        """
        pass

    # ###########################################################################
    # The methods below this point can be ignored by most non-FE developers
    # ###########################################################################

    # noinspection PyMethodMayBeStatic
    def get_fe_models(self) -> Set[Model]:
        """A method to get any models held by this Op.

        All users and most developers can safely ignore this method. This method may be invoked to gather and manipulate
        models, for example by the Network during load_epoch().

        Returns:
            Any models held by this Op.
        """
        return set()

    # noinspection PyMethodMayBeStatic
    def get_fe_loss_keys(self) -> Set[str]:
        """A method to get any loss keys held by this Op.

        All users and most developers can safely ignore this method. This method may be invoked to gather information
        about losses, for example by the Network in get_loss_keys().

        Returns:
            Any loss keys held by this Op.
        """
        return set()

    # noinspection PyMethodMayBeStatic
    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        """A method to get / set whether this Op should retain network gradients after computing them.

        All users and most developers can safely ignore this method. Ops which do not compute gradients should leave
        this method alone. If this method is invoked with `retain` as True or False, then the gradient computations
        performed by this Op should retain or discard the graph respectively afterwards.

        Args:
            retain: If None, then return the current retain_graph status of the Op. If True or False, then set the
                retain_graph status of the op to the new status and return the new status.

        Returns:
            Whether this Op will retain the backward gradient graph after it's forward pass, or None if this Op does not
            compute backward gradients.
        """
        return None


@traceable()
class LambdaOp(TensorOp):
    """An Operator that performs any specified function as forward function.

    Args:
        fn: The function to be executed.
        inputs: Key(s) from which to retrieve data from the data dictionary.
        outputs: Key(s) under which to write the outputs of this Op back to the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 fn: Callable,
                 inputs: Union[None, str, Iterable[str]] = None,
                 outputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.fn = fn
        self.in_list = True

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> Union[Tensor, List[Tensor]]:
        return self.fn(*data)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Union

import numpy as np
from albumentations import Compose, ImageOnlyTransform, ReplayCompose

from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable


@traceable()
class ImageOnlyAlbumentation(NumpyOp):
    """Operators which apply to single images (as opposed to images + masks or images + bounding boxes).

    This is a wrapper for functionality provided by the Albumentations library:
    https://github.com/albumentations-team/albumentations. A useful visualization tool for many of the possible effects
    it provides is available at https://albumentations-demo.herokuapp.com.

    Args:
        func: An Albumentation function to be invoked.
        inputs: Key(s) from which to retrieve data from the data dictionary. If more than one key is provided, the
            `func` will be run in replay mode so that the exact same augmentation is applied to each value.
        outputs: Key(s) under which to write the outputs of this Op back to the data dictionary.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 func: ImageOnlyTransform,
                 inputs: Union[str, List[str]],
                 outputs: Union[str, List[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        assert len(self.inputs) == len(self.outputs), "Input and Output lengths must match"
        self.func = Compose(transforms=[func])
        self.replay_func = ReplayCompose(transforms=[deepcopy(func)])
        self.in_list, self.out_list = True, True

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        results = [self.replay_func(image=data[0]) if len(data) > 1 else self.func(image=data[0])]
        for i in range(1, len(data)):
            results.append(self.replay_func.replay(results[0]['replay'], image=data[i]))
        return [result["image"] for result in results]



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, TypeVar, Union

import matplotlib.backends.backend_agg as plt_backend_agg
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import torch
from matplotlib.gridspec import GridSpec

from fastestimator.util.util import show_image, to_list

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


class ImgData(OrderedDict):
    """A container for image related data.

    This class is intentionally not @traceable.

    This class is useful for automatically laying out collections of images for comparison and visualization.

    ```python
    d = fe.util.ImgData(y=tf.ones((4,)), x=0.5*tf.ones((4, 32, 32, 3)))
    fig = d.paint_figure()
    plt.show()

    img = 0.5*np.ones((4, 32, 32, 3))
    mask = np.zeros_like(img)
    mask[0, 10:20, 10:30, :] = [1, 0, 0]
    mask[1, 5:15, 5:20, :] = [0, 1, 0]
    bbox = np.array([[[3,7,10,6,'box1'], [20,20,8,8,'box2']]]*4)
    d = fe.util.ImgData(y=tf.ones((4,)), x=[img, mask, bbox])
    fig = d.paint_figure()
    plt.show()
    ```

    Args:
        colormap: What colormap to use when rendering greyscale images. A good colorization option is 'inferno'.
        **kwargs: image_title / image pairs for visualization. Images with the same batch dimensions will be laid out
            side-by-side, with earlier kwargs entries displayed further to the left. The value part of the key/value
            pair can be a list of tensors, in which case the elements of the list are overlaid. This can be useful for
            displaying masks and bounding boxes on top of images. In such cases, the largest image should be put as the
            first entry in the list. Bounding boxes should be shaped like (batch, n_boxes, box), where each box is
            formatted like (x0, y0, width, height[, label]).

    Raises:
        AssertionError: If a list of Tensors is provided as an input, but that list has an inconsistent batch dimension.
    """
    n_elements: Dict[int, List[str]]

    def __init__(self, colormap: str = "Greys", **kwargs: Union[Tensor, List[Tensor]]) -> None:
        self.n_elements = {}  # Not a default dict b/c that complicates the computations later
        self.colormap = colormap
        super().__init__(**kwargs)

    def __setitem__(self, key: str, value: Union[Tensor, List[Tensor]]):
        # Convert all values into a list for consistency
        value = to_list(value)
        batch_size = value[0].shape[0]
        for elem in value[1:]:
            assert elem.shape[0] == batch_size, "Provided item has an inconsistent batch size"
        super().__setitem__(key, value)
        self.n_elements.setdefault(batch_size, []).append(key)

    def __delitem__(self, key: str):
        super().__delitem__(key)
        for k, lst in self.n_elements.items():
            lst.remove(key)
            if len(lst) == 0:
                del self.n_elements[k]

    def _to_grid(self) -> List[List[Tuple[str, np.ndarray]]]:
        """Convert the elements of ImgData into a grid view.

        One row in the grid is generated for each unique batch dimension present within the ImgData. Each column in the
        grid is a tensor with batch dimension matching the current row. Columns are given in the order they were input
        into the ImgData constructor.

        Returns:
            The ImgData arranged as a grid, with entries in the grid as (key, value) pairs.
        """
        sorted_sections = sorted(self.n_elements.keys())
        return [[(key, self[key]) for key in self.n_elements[n_rows]] for n_rows in sorted_sections]

    def _n_rows(self) -> int:
        """Computes how many rows are present in the ImgData grid.

        Returns:
            The number of rows in the ImgData grid.
        """
        return len(self.n_elements)

    def _n_cols(self) -> int:
        """Computes how many columns are present in the ImgData grid.

        Returns:
            The number of columns in the ImgData grid.
        """
        return max((len(elem) for elem in self.n_elements.values()))

    @staticmethod
    def _shape_to_width(shape: Tuple[int], min_width=200) -> int:
        """Decide the width of an image for visualization.

        Args:
            shape: The shape of the image.
            min_width: The minimum desired width for visualization.

        Returns:
            The maximum between the width specified by `shape` and the given `min_width` value.
        """
        if len(shape) < 2:
            # text field, use default width
            pass
        elif len(shape) == 2:
            # image field: width x height
            min_width = max(shape[0], min_width)
        else:
            # image field: batch x width x height
            min_width = max(shape[1], min_width)
        return min_width

    @staticmethod
    def _shape_to_height(shape: Tuple[int], min_height=200) -> int:
        """Decide the height of an image for visualization.

        Args:
            shape: The shape of the image.
            min_height: The minimum desired width for visualization.

        Returns:
            The maximum between the height specified by `shape` and the given `min_height` value.
        """
        if len(shape) < 2:
            # text field, use default width
            pass
        elif len(shape) == 2:
            # image field: width x height
            min_height = max(shape[0], min_height)
        else:
            # image field: batch x width x height
            min_height = max(shape[1], min_height) * shape[0]
        return min_height

    def _widths(self, row: int, gap: int = 50, min_width: int = 200) -> List[Tuple[int, int]]:
        """Get the display widths of a particular row.

        Args:
            row: The row to measure.
            gap: How much space to allow between each column.
            min_width: The minimum width for a column.

        Returns:
            A list of (x1, x2) coordinates marking the beginning and end coordinates of each column in the `row`.
        """
        keys = list(sorted(self.n_elements.keys()))
        # For overlay values consider the zeroth element for the shape
        row = [self[key][0] for key in self.n_elements[keys[row]]]
        widths = [(0, ImgData._shape_to_width(row[0].shape, min_width=min_width))]
        for img in row[1:]:
            widths.append(
                (widths[-1][1] + gap, widths[-1][1] + gap + ImgData._shape_to_width(img.shape, min_width=min_width)))
        return widths

    def _total_width(self, gap: int = 50, min_width: int = 200) -> int:
        """Get the total width necessary for the image by considering the widths of each row.

        Args:
            gap: The horizontal space between each column in the grid.
            min_width: The minimum width of a column.

        Returns:
            The total width of the image.
        """
        return max(
            (self._widths(row, gap=gap, min_width=min_width)[-1][-1] for row in range(len(self.n_elements.keys()))))

    def _heights(self, gap: int = 100, min_height: int = 200) -> List[Tuple[int, int]]:
        """Get the display heights of each row.

        Args:
            gap: How much space to allow between each row.
            min_height: The minimum height for a row.

        Returns:
            A list of (y1, y2) coordinates marking the top and bottom coordinates of each row in the grid.
        """
        keys = list(sorted(self.n_elements.keys()))
        # For overlay values consider the zeroth element for the shape
        rows = [[self[key][0] for key in self.n_elements[keys[row]]] for row in range(self._n_rows())]
        heights = [
            max((ImgData._shape_to_height(elem.shape, min_height=min_height) for elem in rows[i]))
            for i in range(self._n_rows())
        ]
        offset = 10
        result = [(offset, heights[0] + offset)]
        for height in heights[1:]:
            result.append((result[-1][1] + gap, result[-1][1] + gap + height))
        return result

    def _total_height(self, gap: int = 100, min_height: int = 200) -> int:
        """Get the total height necessary for the image by considering the heights of each row.

        Args:
            gap: The vertical space between each row in the grid.
            min_height: The minimum height of a row.

        Returns:
            The total height of the image.
        """
        heights = self._heights(gap=gap, min_height=min_height)
        # Add some space at the top for the labels
        return heights[-1][1] + 30

    def _batch_size(self, row: int) -> int:
        """Get the batch size associated with the given `row`.

        Args:
            row: The row for which to report the batch size.

        Returns:
            The batch size of all of the entries in the row.
        """
        return sorted(self.n_elements.keys())[row]

    def paint_figure(self,
                     height_gap: int = 100,
                     min_height: int = 200,
                     width_gap: int = 50,
                     min_width: int = 200,
                     dpi: int = 96,
                     save_path: Optional[str] = None) -> plt.Figure:
        """Visualize the current ImgData entries in a matplotlib figure.

        ```python
        d = fe.util.ImgData(y=tf.ones((4,)), x=0.5*tf.ones((4, 32, 32, 3)))
        fig = d.paint_figure()
        plt.show()
        ```

        Args:
            height_gap: How much space to put between each row.
            min_height: The minimum height of a row.
            width_gap: How much space to put between each column.
            min_width: The minimum width of a column.
            dpi: The resolution of the image to display.
            save_path: If provided, the figure will be saved to the given path.

        Returns:
            The handle to the generated matplotlib figure.
        """
        total_width = self._total_width(gap=width_gap, min_width=min_width)
        total_height = self._total_height(gap=height_gap, min_height=min_height)

        fig = plt.figure(figsize=(total_width / dpi, total_height / dpi), dpi=dpi)

        grid = self._to_grid()
        # TODO - elements with batch size = 1 should be laid out in a grid like for plotting
        for row_idx, (start_height, end_height) in enumerate(self._heights(gap=height_gap, min_height=min_height)):
            row = grid[row_idx]
            batch_size = self._batch_size(row_idx)
            gs = GridSpec(nrows=batch_size,
                          ncols=total_width,
                          figure=fig,
                          left=0.0,
                          right=1.0,
                          bottom=start_height / total_height,
                          top=end_height / total_height,
                          hspace=0.05,
                          wspace=0.0)
            for batch_idx in range(batch_size):
                for col_idx, width in enumerate(self._widths(row=row_idx, gap=width_gap, min_width=min_width)):
                    ax = fig.add_subplot(gs[batch_idx, width[0]:width[1]])
                    img_stack = [elem[batch_idx] for elem in row[col_idx][1]]
                    for idx, img in enumerate(img_stack):
                        show_image(img,
                                   axis=ax,
                                   fig=fig,
                                   title=row[col_idx][0] if (batch_idx == 0 and idx == 0) else None,
                                   stack_depth=idx,
                                   color_map=self.colormap)
        if save_path:
            plt.savefig(save_path, dpi=dpi, bbox_inches="tight")
        return fig

    def paint_numpy(self,
                    height_gap: int = 100,
                    min_height: int = 200,
                    width_gap: int = 50,
                    min_width: int = 200,
                    dpi: int = 96) -> np.ndarray:
        """Visualize the current ImgData entries into an image stored in a numpy array.

        ```python
        d = fe.util.ImgData(y=tf.ones((4,)), x=0.5*tf.ones((4, 32, 32, 3)))
        img = d.paint_numpy()
        plt.imshow(img[0])
        plt.show()
        ```

        Args:
            height_gap: How much space to put between each row.
            min_height: The minimum height of a row.
            width_gap: How much space to put between each column.
            min_width: The minimum width of a column.
            dpi: The resolution of the image to display.

        Returns:
            A numpy array with dimensions (1, height, width, 3) containing an image representation of this ImgData.
        """
        fig = self.paint_figure(height_gap=height_gap,
                                min_height=min_height,
                                width_gap=width_gap,
                                min_width=min_width,
                                dpi=dpi)
        canvas = plt_backend_agg.FigureCanvasAgg(fig)
        canvas.draw()
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        w, h = fig.canvas.get_width_height()
        data = data.reshape([h, w, 4])[:, :, 0:3]
        plt.close(fig)
        return np.stack([data])  # Add a batch dimension



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import csv
import os
import shutil
import tempfile
import unittest
from io import StringIO
from unittest.mock import patch

import numpy as np
import tensorflow as tf
import torch
from PIL import Image

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal, sample_system_object, sample_system_object_torch
from fastestimator.trace.io import TensorBoard
from fastestimator.trace.io.tensorboard import _TfWriter, _TorchWriter
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import FeInputSpec


def getfilepath():
    path = os.path.join(*[tempfile.gettempdir(), 'tensorboard', 'train'])
    for filename in os.listdir(path):
        return os.path.join(path, filename)


class TestTensorboard(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tf_data = Data({
            'x': tf.random.normal(shape=(1, 28, 28, 3)),
            'y': tf.random.uniform(shape=(1, ), maxval=10, dtype=tf.int32),
            'images': tf.random.normal(shape=(1, 28, 28, 3)),
            'embed': np.ones(shape=(1, 3, 3, 3)),
            'embed_images': np.ones(shape=(1, 3, 3, 3))
        })
        cls.torch_data = Data({
            'x': torch.rand(size=(1, 1, 28, 28)),
            'y': torch.rand(size=(3, )),
            'images': torch.rand(size=(1, 3, 28, 28)),
            'embed': np.ones(shape=(1, 3, 3, 3)),
            'embed_images': np.ones(shape=(1, 3, 3, 3))
        })
        cls.log_dir = os.path.join(tempfile.gettempdir(), 'tensorboard')
        cls.train_path = os.path.join(cls.log_dir, 'train')
        cls.embed_path = os.path.join(cls.log_dir, 'train', '00001', 'embed')
        cls.on_begin_msg = "FastEstimator-Tensorboard: writing logs to {}".format(cls.log_dir)

    def test_tf_on_begin(self):
        tensorboard = TensorBoard(log_dir=self.log_dir)
        tensorboard.system = sample_system_object()
        tensorboard.system.global_step = 1
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            tensorboard.on_begin(data=self.tf_data)
            log = fake_stdout.getvalue().strip()
            self.assertEqual(log, self.on_begin_msg)

    def test_tf_on_batch_end(self):
        tensorboard = TensorBoard(log_dir=self.log_dir, weight_histogram_freq=1, update_freq=1)
        tensorboard.system = sample_system_object()
        tensorboard.system.global_step = 1
        tensorboard.writer = _TfWriter(self.log_dir, '', tensorboard.system.network)
        model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam')
        tensorboard.system.network.epoch_models = {model}
        if os.path.exists(self.train_path):
            shutil.rmtree(self.train_path)
        tensorboard.on_batch_end(data=self.tf_data)
        filepath = getfilepath()
        for e in tf.compat.v1.train.summary_iterator(filepath):
            for v in e.summary.value:
                if v.tag == "tf_dense_1/bias_0":
                    output = v.histo.num
                    self.assertEqual(output, 10.0)

    def test_tf_on_epoch_end(self):
        tensorboard = TensorBoard(log_dir=self.log_dir,
                                  weight_histogram_freq=1,
                                  update_freq=1,
                                  write_images='images',
                                  write_embeddings='embed',
                                  embedding_images='embed_images')
        tensorboard.system = sample_system_object()
        tensorboard.system.global_step = 1
        tensorboard.writer = _TfWriter(self.log_dir, '', tensorboard.system.network)
        model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam')
        tensorboard.system.network.epoch_models = {model}
        if os.path.exists(self.train_path):
            shutil.rmtree(self.train_path)
        tensorboard.on_epoch_end(data=self.tf_data)
        tsv_path = os.path.join(self.embed_path, 'tensors.tsv')
        embed_img_path = os.path.join(self.embed_path, 'sprite.png')
        # get tensor data from tsv file
        fo = open(tsv_path)
        tsv_content = csv.reader(fo, delimiter='\t')
        for row in tsv_content:
            tsv_data = row
        fo.close()
        # get the image data
        output_img = np.asarray(Image.open(embed_img_path))
        with self.subTest('Check if tensors.tsv was generated'):
            self.assertTrue(os.path.exists(tsv_path))
        with self.subTest('Check if embed image was generated'):
            self.assertTrue(os.path.exists(embed_img_path))
        with self.subTest('Check content of tensors.tsv'):
            self.assertEqual(tsv_data, 27 * ['1.0'])
        with self.subTest('Check embed image content'):
            self.assertTrue(is_equal(output_img, 255 * np.ones(shape=(3, 3, 3), dtype=np.int)))

    def test_torch_on_begin(self):
        tensorboard = TensorBoard(log_dir=self.log_dir)
        tensorboard.system = sample_system_object_torch()
        tensorboard.system.global_step = 1
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            tensorboard.on_begin(data=self.torch_data)
            log = fake_stdout.getvalue().strip()
            self.assertEqual(log, self.on_begin_msg)

    def test_torch_on_batch_end(self):
        tensorboard = TensorBoard(log_dir=self.log_dir, weight_histogram_freq=1, update_freq=1)
        tensorboard.system = sample_system_object_torch()
        tensorboard.system.global_step = 1
        tensorboard.writer = _TorchWriter(self.log_dir, '', tensorboard.system.network)
        model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='torch')
        model.fe_input_spec = FeInputSpec(self.torch_data['x'], model)
        tensorboard.system.network.epoch_models = {model}
        if os.path.exists(self.train_path):
            shutil.rmtree(self.train_path)
        tensorboard.on_batch_end(data=self.torch_data)
        filepath = getfilepath()
        for e in tf.compat.v1.train.summary_iterator(filepath):
            for v in e.summary.value:
                if v.tag == "torch_fc1/bias":
                    output = v.histo.num
                    self.assertEqual(output, 64.0)

    def test_torch_on_epoch_end(self):
        tensorboard = TensorBoard(log_dir=self.log_dir,
                                  weight_histogram_freq=1,
                                  update_freq=1,
                                  write_images='images',
                                  write_embeddings='embed',
                                  embedding_images='embed_images')
        tensorboard.system = sample_system_object_torch()
        tensorboard.system.global_step = 1
        tensorboard.writer = _TorchWriter(self.log_dir, '', tensorboard.system.network)
        model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam')
        tensorboard.system.network.epoch_models = {model}
        if os.path.exists(self.train_path):
            shutil.rmtree(self.train_path)
        tensorboard.on_epoch_end(data=self.torch_data)
        tsv_path = os.path.join(self.embed_path, 'tensors.tsv')
        embed_img_path = os.path.join(self.embed_path, 'sprite.png')
        # get tensor data from tsv file
        fo = open(tsv_path)
        tsv_content = csv.reader(fo, delimiter='\t')
        for row in tsv_content:
            tsv_data = row
        fo.close()
        # get the image data
        output_img = np.asarray(Image.open(embed_img_path))
        with self.subTest('Check if tensors.tsv was generated'):
            self.assertTrue(os.path.exists(tsv_path))
        with self.subTest('Check if embed image was generated'):
            self.assertTrue(os.path.exists(embed_img_path))
        with self.subTest('Check content of tensors.tsv'):
            self.assertEqual(tsv_data, 27 * ['1.0'])
        with self.subTest('Check embed image content'):
            self.assertTrue(is_equal(output_img, 255 * np.ones(shape=(3, 3, 3), dtype=np.int)))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestClipByValue(unittest.TestCase):
    def test_clip_by_value_tf_input(self):
        t = tf.constant([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=-2, max_value=3)
        obj2 = tf.constant([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_torch_input(self):
        t = torch.tensor([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=-2, max_value=3)
        obj2 = torch.tensor([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_np_input(self):
        t = np.array([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=-2, max_value=3)
        obj2 = np.array([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_tensor_min_max_value_tf(self):
        t = tf.constant([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=tf.constant(-2), max_value=3)
        obj2 = tf.constant([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_tensor_min_max_value_list_tf(self):
        t = tf.constant([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=tf.constant([-2]), max_value=3)
        obj2 = tf.constant([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_tensor_min_max_value_torch(self):
        t = torch.tensor([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=-2, max_value=torch.tensor(3))
        obj2 = torch.tensor([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))

    def test_clip_by_value_tensor_min_max_value_list_torch(self):
        t = torch.tensor([-5, 4, 2, 0, 9, -2])
        obj1 = fe.backend.clip_by_value(t, min_value=-2, max_value=torch.tensor([3]))
        obj2 = torch.tensor([-2, 3, 2, 0, 3, -2])
        self.assertTrue(is_equal(obj1, obj2))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Tuple, Union

from albumentations.augmentations.transforms import RandomSunFlare as RandomSunFlareAlb

from fastestimator.op.numpyop.univariate.univariate import ImageOnlyAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class RandomSunFlare(ImageOnlyAlbumentation):
    """Add a sun flare to the image.

    Args:
        inputs: Key(s) of images to be normalized.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        flare_roi: region of the image where flare will appear (x_min, y_min, x_max, y_max). All values should be
            in range [0, 1].
        angle_lower: should be in range [0, `angle_upper`].
        angle_upper: should be in range [`angle_lower`, 1].
        num_flare_circles_lower: lower limit for the number of flare circles.
            Should be in range [0, `num_flare_circles_upper`].
        num_flare_circles_upper: upper limit for the number of flare circles.
            Should be in range [`num_flare_circles_lower`, inf].
        src_radius: Radius of the flare.
        src_color: Color of the flare (R,G,B).

    Image types:
        uint8, float32
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 flare_roi: Tuple[float, float, float, float] = (0, 0, 1, 0.5),
                 angle_lower: float = 0.0,
                 angle_upper: float = 1.0,
                 num_flare_circles_lower: int = 6,
                 num_flare_circles_upper: int = 10,
                 src_radius: int = 400,
                 src_color: Tuple[int, int, int] = (255, 255, 255)):
        super().__init__(
            RandomSunFlareAlb(flare_roi=flare_roi,
                              angle_lower=angle_lower,
                              angle_upper=angle_upper,
                              num_flare_circles_lower=num_flare_circles_lower,
                              num_flare_circles_upper=num_flare_circles_upper,
                              src_radius=src_radius,
                              src_color=src_color,
                              always_apply=True),
            inputs=inputs,
            outputs=outputs,
            mode=mode,
            ds_id=ds_id)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestMaximum(unittest.TestCase):
    def test_maximum_np_input(self):
        n1 = np.array([[2, 7, 6]])
        n2 = np.array([[2, 7, 5]])
        res = fe.backend.maximum(n1, n2)
        self.assertTrue(is_equal(res, n1))

    def test_maximum_tf_input(self):
        t1 = tf.constant([[2, 7, 6]])
        t2 = tf.constant([[2, 7, 5]])
        res = fe.backend.maximum(t1, t2)
        self.assertTrue(is_equal(res, t1))

    def test_maximum_torch_input(self):
        p1 = torch.tensor([[2, 7, 6]])
        p2 = torch.tensor([[2, 7, 5]])
        res = fe.backend.maximum(p1, p2)
        self.assertTrue(is_equal(res, p1))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# see a complete list of effects in albumentation page: https://albumentations-demo.herokuapp.com
from fastestimator.op.numpyop.univariate.autocontrast import AutoContrast
from fastestimator.op.numpyop.univariate.binarize import Binarize
from fastestimator.op.numpyop.univariate.blur import Blur
from fastestimator.op.numpyop.univariate.brightness import Brightness
from fastestimator.op.numpyop.univariate.calibate import Calibrate
from fastestimator.op.numpyop.univariate.channel_dropout import ChannelDropout
from fastestimator.op.numpyop.univariate.channel_shuffle import ChannelShuffle
from fastestimator.op.numpyop.univariate.channel_transpose import ChannelTranspose
from fastestimator.op.numpyop.univariate.clahe import CLAHE
from fastestimator.op.numpyop.univariate.coarse_dropout import CoarseDropout
from fastestimator.op.numpyop.univariate.color import Color
from fastestimator.op.numpyop.univariate.color_jitter import ColorJitter
from fastestimator.op.numpyop.univariate.contrast import Contrast
from fastestimator.op.numpyop.univariate.downscale import Downscale
from fastestimator.op.numpyop.univariate.equalize import Equalize
from fastestimator.op.numpyop.univariate.expand_dims import ExpandDims
from fastestimator.op.numpyop.univariate.from_float import FromFloat
from fastestimator.op.numpyop.univariate.gaussian_blur import GaussianBlur
from fastestimator.op.numpyop.univariate.gaussian_noise import GaussianNoise
from fastestimator.op.numpyop.univariate.hadamard import Hadamard
from fastestimator.op.numpyop.univariate.hue_saturation_value import HueSaturationValue
from fastestimator.op.numpyop.univariate.iaa_additive_gaussian_noise import IAAAdditiveGaussianNoise
from fastestimator.op.numpyop.univariate.image_compression import ImageCompression
from fastestimator.op.numpyop.univariate.invert_img import InvertImg
from fastestimator.op.numpyop.univariate.iso_noise import ISONoise
from fastestimator.op.numpyop.univariate.median_blur import MedianBlur
from fastestimator.op.numpyop.univariate.minmax import Minmax
from fastestimator.op.numpyop.univariate.motion_blur import MotionBlur
from fastestimator.op.numpyop.univariate.multiplicative_noise import MultiplicativeNoise
from fastestimator.op.numpyop.univariate.normalize import Normalize
from fastestimator.op.numpyop.univariate.onehot import Onehot
from fastestimator.op.numpyop.univariate.pad_sequence import PadSequence
from fastestimator.op.numpyop.univariate.posterize import Posterize
from fastestimator.op.numpyop.univariate.random_brightness_contrast import RandomBrightnessContrast
from fastestimator.op.numpyop.univariate.random_fog import RandomFog
from fastestimator.op.numpyop.univariate.random_gamma import RandomGamma
from fastestimator.op.numpyop.univariate.random_rain import RandomRain
from fastestimator.op.numpyop.univariate.random_shadow import RandomShadow
from fastestimator.op.numpyop.univariate.random_shapes import RandomShapes
from fastestimator.op.numpyop.univariate.random_snow import RandomSnow
from fastestimator.op.numpyop.univariate.random_sun_flare import RandomSunFlare
from fastestimator.op.numpyop.univariate.read_image import ReadImage
from fastestimator.op.numpyop.univariate.reshape import Reshape
from fastestimator.op.numpyop.univariate.rgb_shift import RGBShift
from fastestimator.op.numpyop.univariate.rua import RUA
from fastestimator.op.numpyop.univariate.sharpness import Sharpness
from fastestimator.op.numpyop.univariate.shear_x import ShearX
from fastestimator.op.numpyop.univariate.shear_y import ShearY
from fastestimator.op.numpyop.univariate.solarize import Solarize
from fastestimator.op.numpyop.univariate.to_array import ToArray
from fastestimator.op.numpyop.univariate.to_float import ToFloat
from fastestimator.op.numpyop.univariate.to_gray import ToGray
from fastestimator.op.numpyop.univariate.to_sepia import ToSepia
from fastestimator.op.numpyop.univariate.tokenize import Tokenize
from fastestimator.op.numpyop.univariate.translate_x import TranslateX
from fastestimator.op.numpyop.univariate.translate_y import TranslateY
from fastestimator.op.numpyop.univariate.word_to_id import WordtoId



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

from fastestimator.op.tensorop.gradient import Watch
from fastestimator.test.unittest_util import is_equal


class TestWatch(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tf_data = tf.Variable([1., 2., 4.])
        cls.tf_output = [tf.Variable([1., 2., 4.], dtype=tf.float32), tf.constant([1, 4, 16], dtype=tf.float32)]
        cls.torch_data = torch.tensor([1., 2., 4.], requires_grad=True)
        cls.torch_output = [torch.tensor([1, 2, 4], dtype=torch.float32), torch.tensor([1, 4, 16], dtype=torch.float32)]

    def test_tf_input(self):
        def run_watch():
            with tf.GradientTape(persistent=True) as tape:
                x = self.tf_data * self.tf_data
                output = watch.forward(data=[self.tf_data, x], state={'tape': tape})
                self.assertTrue(is_equal(output, self.tf_output))
        watch = Watch(inputs='x')
        strategy = tf.distribute.get_strategy()
        if isinstance(strategy, tf.distribute.MirroredStrategy):
            strategy.run(run_watch)
        else:
            run_watch()

    def test_torch_input(self):
        watch = Watch(inputs='x')
        x = self.torch_data * self.torch_data
        output = watch.forward(data=[self.torch_data, x], state={'tape': None})
        self.assertTrue(is_equal(output, self.torch_output))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Tuple, Union

import numpy as np

from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable


@traceable()
class Reshape(NumpyOp):
    """An Op which re-shapes data to a target shape.

    Args:
        shape: The desired output shape. At most one value may be -1 to put all of the leftover elements into that axis.
        inputs: Key(s) of the data to be reshaped.
        outputs: Key(s) into which to write the converted data.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 shape: Union[int, Tuple[int, ...]],
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.shape = shape
        self.in_list, self.out_list = True, True

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        return [self._apply_reshape(elem) for elem in data]

    def _apply_reshape(self, data):
        data = np.reshape(data, self.shape)
        return data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.op.tensorop.loss.cross_entropy import CrossEntropy
from fastestimator.op.tensorop.loss.mix_loss import MixLoss


class TestMixUpBatch(unittest.TestCase):
    def test_mixup_batch_tf(self):
        true_binary = tf.constant([[1.0], [2.0], [4.0]])
        pred_binary = tf.constant([[1.0], [3.0], [4.5]])

        ml = MixLoss(CrossEntropy(inputs=("y_pred", "y"), mode="train", outputs="loss"), lam="lambda")
        output = ml.forward(data=[0.1, pred_binary, true_binary], state={})

        self.assertTrue(np.allclose(output.numpy(), -20.444319))

    def test_mixup_batch_torch(self):
        true_binary = torch.tensor([[1], [0], [1], [0]])
        pred_binary = torch.tensor([[0.9], [0.3], [0.8], [0.1]])

        ml = MixLoss(CrossEntropy(inputs=("y_pred", "y"), mode="train", outputs="loss"), lam="lambda")
        output = ml.forward(data=[0.1, pred_binary, true_binary], state={})

        self.assertTrue(np.allclose(output.detach().numpy(), 1.6889441))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Union

from albumentations.augmentations.transforms import RandomSnow as RandomSnowAlb

from fastestimator.op.numpyop.univariate.univariate import ImageOnlyAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class RandomSnow(ImageOnlyAlbumentation):
    """Bleach out some pixels to simulate snow.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        snow_point_lower: Lower bound of the amount of snow. Should be in the range [0, 1].
        snow_point_upper: Upper bound of the amount of snow. Should be in the range [0, 1].
        brightness_coeff: A larger number will lead to a more snow on the image. Should be >= 0.

    Image types:
        uint8, float32
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 snow_point_lower: float = 0.1,
                 snow_point_upper: float = 0.3,
                 brightness_coeff: float = 2.5):
        super().__init__(
            RandomSnowAlb(snow_point_lower=snow_point_lower,
                          snow_point_upper=snow_point_upper,
                          brightness_coeff=brightness_coeff,
                          always_apply=True),
            inputs=inputs,
            outputs=outputs,
            mode=mode,
            ds_id=ds_id)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.schedule import EpochScheduler


class TestEpochScheduler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.input_data = {1: "a", 3: "b", 4: None, 100: "c"}
        cls.values = ['a', 'b', None, 'c']
        cls.scheduler = EpochScheduler(cls.input_data)

    def test_get_current_value(self):
        self.assertEqual(self.scheduler.get_current_value(2), 'a')

    def test_get_all_values(self):
        self.assertEqual(self.scheduler.get_all_values(), self.values)

    def test_get_last_key(self):
        self.assertEqual(self.scheduler._get_last_key(3), 3)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import sys
from typing import Any, Dict, List

from fastestimator.summary.logs import parse_log_dir
from fastestimator.util.cli_util import SaveAction


def logs(args: Dict[str, Any], unknown: List[str]) -> None:
    """A method to invoke the FE logging function using CLI-provided arguments.

    Args:
        args: The arguments to be fed to the parse_log_dir() method.
        unknown: Any cli arguments not matching known inputs for the parse_log_dir() method.

    Raises:
        SystemExit: If `unknown` arguments were provided by the user.
    """
    if len(unknown) > 0:
        print("error: unrecognized arguments: ", str.join(", ", unknown))
        sys.exit(-1)
    group_by = args['group_by']
    if isinstance(group_by, list):
        group_by = group_by[0]
    if group_by == '_n':
        group_by = r'(.*)_[\d]+' + '\\' + args['extension']
    parse_log_dir(args['log_dir'],
                  args['extension'],
                  args['recursive'],
                  args['smooth'],
                  args['save'],
                  args['save_dir'],
                  args['ignore'],
                  args['include'],
                  args['share_legend'],
                  args['pretty_names'],
                  group_by)


def configure_log_parser(subparsers: argparse._SubParsersAction) -> None:
    """Add a logging parser to an existing argparser.

    Args:
        subparsers: The parser object to be appended to.
    """
    parser = subparsers.add_parser('logs',
                                   description='Generates comparison graphs amongst one or more log files',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                   allow_abbrev=False)
    parser.add_argument('log_dir',
                        metavar='<Log Dir>',
                        type=str,
                        help="The path to a folder containing one or more log files")
    parser.add_argument('--extension',
                        metavar='E',
                        type=str,
                        help="The file type / extension of your logs",
                        default=".txt")
    parser.add_argument('--recursive', action='store_true', help="Recursively search sub-directories for log files")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--ignore',
                       metavar='I',
                       type=str,
                       nargs='+',
                       help="The names of metrics to ignore though they may be present in the log files")
    group.add_argument('--include',
                       metavar='Y',
                       type=str,
                       nargs='+',
                       help="The names of metrics to include. If provided, any other metrics will be ignored.")
    parser.add_argument('--smooth',
                        metavar='<float>',
                        type=float,
                        help="The amount of gaussian smoothing to apply (zero for no smoothing)",
                        default=1)
    parser.add_argument('--pretty_names', help="Clean up the metric names for display", action='store_true')
    parser.add_argument('--group_by',
                        metavar='G',
                        type=str,
                        nargs=1,
                        help="A regex pattern to group different logs together and display their mean+-stdev. For "
                             r"example, you could use --G '(.*)_[\d]+\.txt' to group files of the form "
                             "<name>_<number>.txt by their <name>. We anticipate this being the common usecase, so you "
                             "can use --G _n as a shortcut for that functionality.")

    legend_group = parser.add_argument_group('legend arguments')
    legend_x_group = legend_group.add_mutually_exclusive_group(required=False)
    legend_x_group.add_argument('--common_legend',
                                dest='share_legend',
                                help="Generate one legend total",
                                action='store_true',
                                default=True)
    legend_x_group.add_argument('--split_legend',
                                dest='share_legend',
                                help="Generate one legend per graph",
                                action='store_false',
                                default=False)

    save_group = parser.add_argument_group('output arguments')
    save_x_group = save_group.add_mutually_exclusive_group(required=False)
    save_x_group.add_argument(
        '--save',
        nargs='?',
        metavar='<Save Dir>',
        dest='save',
        action=SaveAction,
        default=False,
        help="Save the output image. May be accompanied by a directory into \
                                              which the file is saved. If no output directory is specified, the log \
                                              directory will be used")
    save_x_group.add_argument('--display',
                              dest='save',
                              action='store_false',
                              help="Render the image to the UI (rather than saving it)",
                              default=True)
    save_x_group.set_defaults(save_dir=None)
    parser.set_defaults(func=logs)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.get_gradient import get_gradient
from fastestimator.backend.reduce_mean import reduce_mean
from fastestimator.backend.update_model import update_model
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_set

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class UpdateOp(TensorOp):
    """This class performs updates to a model's weights based on the loss.

    Args:
        model: Model instance compiled by fe.build.
        loss_name: The input loss key.
        gradients: An optional key containing model gradients. These will be directly applied to the model weights
            during an update. If not provided, gradients will be computed based on the specified loss_name, which will
            automatically handle any desired mixed-precision scaling. This argument shouldn't be used if mixed-precision
            training is enabled.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        merge_grad: The gradient accumulation times before model update. Ex: if `merge_grad` = 3, for every three Op
            calls only the third one updates the model. The first two calls only accumulate its gradients. This default
            value is 1 and it will update the model at every step.
        defer: Whether to defer the actual application of the update until the end of the step. This can be necessary
            in PyTorch when trying to update multiple models which depend on one another (ex. certain GANs). By default,
            all UpdateOps which appear contiguously as the last ops of a Network will be deferred. We hope that you will
            never need to worry about this flag, but it's here for you if you need it.

    Raise:
        ValueError: When model is mixed-precision and `gradients` is provided.
        ValueError: Network framework is not one of "tf" or "torch".
        ValueError: `merge_grad` is larger than 1 in multi-GPU configuration.
        RuntimeError: If attempting to modify a PyTorch model which relied on gradients within a different PyTorch model
            which has in turn already undergone a non-deferred update.
    """
    def __init__(self,
                 model: Union[tf.keras.Model, torch.nn.Module],
                 loss_name: str,
                 gradients: Optional[str] = None,
                 mode: Union[None, str, Iterable[str]] = "train",
                 ds_id: Union[None, str, Iterable[str]] = None,
                 merge_grad: int = 1,
                 defer: bool = False):
        self.extra_loss = isinstance(model, tf.keras.Model) and model.losses
        if gradients is None:
            super().__init__(inputs=loss_name, outputs=None, mode=mode, ds_id=ds_id)
        else:
            if model.mixed_precision:
                raise ValueError("Mixed precision training cannot take input gradients, because the gradients need to "
                                 "be computed in this module")
            if self.extra_loss:
                print("FastEstimator-Warn: Extra model losses are detected and they will be ignored since the gradients"
                      " are not computed in UpdateOp class.")
            super().__init__(inputs=gradients, outputs=None, mode=mode, ds_id=ds_id)

        if torch.cuda.device_count() > 1 and merge_grad > 1:
            raise ValueError("Currently FastEstimator doesn't support merge_grad feature in multi-GPU configuration "
                             "and thus 'merge_grad' cannot be larger than 1")

        if not hasattr(model, "loss_name"):
            model.loss_name = {loss_name}
        else:
            model.loss_name.add(loss_name)

        self.model = model
        self.retain_graph = False
        self.defer = defer
        self.gradients = gradients
        self.loss_name = loss_name
        self.merge_grad = merge_grad
        self.framework = None

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        if framework not in ["tf", "torch"]:
            raise ValueError(f"Unrecognized framework {framework}")

        self.framework = framework

        if self.merge_grad > 1:
            if framework == "tf":
                self.step = tf.Variable(0, trainable=False, dtype=tf.int64)
                self.grad_sum = [tf.Variable(tf.zeros_like(x), trainable=False) for x in self.model.trainable_variables]
            else:  # framework == "torch"
                self.step = torch.tensor(0, dtype=torch.int64).to(device)
                self.grad_sum = [torch.zeros_like(x).to(device) for x in self.model.parameters() if x.requires_grad]

    def get_fe_models(self) -> Set[Model]:
        return {self.model}

    def get_fe_loss_keys(self) -> Set[str]:
        return to_set(self.loss_name)

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        if retain is not None:
            self.retain_graph = retain
        return self.retain_graph

    def forward(self, data: Union[Tensor, List[Tensor]], state: Dict[str, Any]) -> None:
        if state["warmup"]:
            return

        if self.gradients is None:  # data is loss
            loss = self._loss_preprocess(data)
            gradients = self._get_gradient(loss, state["tape"])
        else:  # data is gradients
            gradients = data
        gradients = self._gradient_postprocess(gradients)

        if self.merge_grad > 1:
            self._merge_grad_update(gradients, deferred=state["deferred"])
        else:
            update_model(model=self.model, gradients=gradients, defer=self.defer, deferred=state["deferred"])

    def _loss_preprocess(self, loss: Union[Tensor, List[Tensor]]) -> Union[Tensor, List[Tensor]]:
        """Loss preprocess for multi-GPU and mixed-precision training.

        Args:
            loss: Unprocessed loss.

        Returns:
            Processed loss.
        """
        if self.extra_loss:
            loss = loss + tf.reduce_sum(self.model.losses)
        loss = reduce_mean(loss)

        if self.framework == "tf":
            # scale up loss for mixed precision training to avoid underflow
            if self.model.mixed_precision:
                loss = self.model.current_optimizer.get_scaled_loss(loss)
            # for multi-gpu training, the gradient will be combined by sum, normalize the loss
            strategy = tf.distribute.get_strategy()
            if isinstance(strategy, tf.distribute.MirroredStrategy):
                loss = loss / strategy.num_replicas_in_sync

        else:  # self.framework == "torch"
            if self.model.current_optimizer.scaler is not None:
                # scale up loss for mixed precision training to avoid underflow
                loss = self.model.current_optimizer.scaler.scale(loss)

        return loss

    def _get_gradient(self, loss: Union[Tensor, List[Tensor]],
                      tape: Optional[tf.GradientTape] = None) -> Union[Tensor, List[Tensor]]:
        """Get gradient from loss with repect to self.model.

        Args:
            loss: Input loss.
            tape: A TensorFlow GradientTape which was recording when the `loss` was computed (iff using TensorFlow).

        Returns:
            Computed gradients.
        """
        if self.framework == "tf":
            gradients = get_gradient(loss, self.model.trainable_variables, tape=tape)

        else:  # self.framework == "torch"
            trainable_params = [p for p in self.model.parameters() if p.requires_grad]
            try:
                gradients = get_gradient(loss, trainable_params, retain_graph=self.retain_graph)
            except RuntimeError as err:
                if err.args and isinstance(err.args[0], str) and err.args[0].startswith(
                        'one of the variables needed for gradient computation has been modified by an inplace operation'
                ):
                    raise RuntimeError(
                        "When computing gradients for '{}', some variables it relied on during the forward pass had"
                        " been updated. Consider setting defer=True in earlier UpdateOps related to models which "
                        "interact with this one.".format(self.model.model_name))
                raise err

        return gradients

    def _gradient_postprocess(self, gradients: Union[Tensor, List[Tensor]]) -> Union[Tensor, List[Tensor]]:
        """Gradient postprocess for multi-GPU and mixed-precision training.

        Args:
            gradients: Unprocessed gradients.

        Returns:
            Processed gradients.
        """
        if self.framework == "tf":
            if self.gradients is not None:  # when user provide gradients
                strategy = tf.distribute.get_strategy()
                # for multi-gpu training, the gradient will be combined by sum, normalize the gradient
                if isinstance(strategy, tf.distribute.MirroredStrategy):
                    gradients = [gs / strategy.num_replicas_in_sync for gs in gradients]

            if self.model.mixed_precision:
                # scale down gradient to balance scale-up loss
                gradients = self.model.current_optimizer.get_unscaled_gradients(gradients)

        return gradients

    def _merge_grad_update(self,
                           gradients: Union[Tensor, List[Tensor]],
                           deferred: Optional[Dict[str, List[Callable[[], None]]]] = None) -> None:
        """Accumulate gradients and update the model at certain frequency of invocation.

        Args:
            gradients: Input gradients.
            deferred: A dictionary in which model update functions are stored.
        """

        # add current gradient to the cumulative gradient
        for gs, g in zip(self.grad_sum, gradients):
            self._assign_add(gs, g)

        self._assign_add(self.step, 1)

        if self.step % self.merge_grad == 0:
            average_grad = [gs / self.merge_grad for gs in self.grad_sum]
            update_model(model=self.model, gradients=average_grad, defer=self.defer, deferred=deferred)
            for gs in self.grad_sum:
                self._assign_add(gs, -gs)  # zero the gradient in place

    def _assign_add(self, a: Tensor, b: Tensor) -> None:
        """In-place addition for both Tensorflow and PyTorch. `a` = `a` + `b`

        Args:
            a: A tensor where in-place addition happens.
            b: Amount to be added.
        """
        if self.framework == "tf":
            a.assign_add(b)
        else:  # self.framework == "torch"
            a += b



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.op.numpyop.multivariate import Crop


class TestCrop(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.height = 10
        cls.width = 10
        cls.single_input = [np.random.rand(28, 28, 3)]
        cls.single_output_shape = (10, 20, 3)
        cls.input_image_and_mask = [np.random.rand(28, 28, 3), np.random.rand(28, 28, 3)]
        cls.image_and_mask_output_shape = (10, 10, 3)

    def test_single_input(self):
        crop = Crop(x_min=0, y_min=10, x_max=20, y_max=20, image_in='x')
        output = crop.forward(data=self.single_input, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.single_output_shape)

    def test_input_image_and_mask(self):
        crop = Crop(x_min=10, y_min=10, x_max=20, y_max=20, image_in='x', mask_in='x_mask')
        output = crop.forward(data=self.input_image_and_mask, state={})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.image_and_mask_output_shape)
        with self.subTest('Check output mask shape'):
            self.assertEqual(output[1].shape, self.image_and_mask_output_shape)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf

from fastestimator.architecture.tensorflow import ResNet9
from fastestimator.architecture.tensorflow.resnet9 import _check_input_size


class TestResNet9(unittest.TestCase):
    def test_resnet9_default(self):
        data = np.ones((1, 32, 32, 3))
        input_data = tf.constant(data)
        model = ResNet9()
        output_shape = model(input_data).numpy().shape
        self.assertEqual(output_shape, (1, 10))

    def test_resnet9_specific_input_size_classes(self):
        size = (16, 16, 1)
        classes = 5
        data = np.ones((1, ) + size)
        input_data = tf.constant(data)
        model = ResNet9(input_size=size, classes=classes)
        output_shape = model(input_data).numpy().shape
        self.assertEqual(output_shape, (1, classes))


class TestCheckInputSize(unittest.TestCase):
    def test_resnet9_check_input_size(self):
        with self.subTest("length not 3"):
            with self.assertRaises(ValueError):
                _check_input_size((1, ))

        with self.subTest("width or height is smaller than 16"):
            with self.assertRaises(ValueError):
                _check_input_size((13, 16, 1))

            with self.assertRaises(ValueError):
                _check_input_size((16, 1, 1))

        with self.subTest("both are not smaller than 16"):
            _check_input_size((16, 16, 1))
            _check_input_size((16, 100, 3))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def permute(tensor: Tensor, permutation: List[int]) -> Tensor:
    """Perform the specified `permutation` on the axes of a given `tensor`.

    This method can be used with Numpy data:
    ```python
    n = np.array([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
    b = fe.backend.permute(n, [2, 0, 1])  # [[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]]
    b = fe.backend.permute(n, [0, 2, 1])  # [[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
    b = fe.backend.permute(t, [2, 0, 1])  # [[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]]
    b = fe.backend.permute(t, [0, 2, 1])  # [[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
    b = fe.backend.permute(p, [2, 0, 1])  # [[[0, 2], [4, 6], [8, 10]], [[1, 3], [5, 7], [9, 11]]]
    b = fe.backend.permute(P, [0, 2, 1])  # [[[0, 2], [1, 3]], [[4, 6], [5, 7]], [[8, 10], [9, 11]]]
    ```

    Args:
        tensor: The tensor to permute.
        permutation: The new axis order to be used. Should be a list containing all integers in range [0, tensor.ndim).

    Returns:
        The `tensor` with axes swapped according to the `permutation`.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        return tf.transpose(tensor, perm=permutation)
    elif isinstance(tensor, torch.Tensor):
        return tensor.permute(*permutation)
    elif isinstance(tensor, np.ndarray):
        return np.transpose(tensor, axes=permutation)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import pickle
from typing import Tuple, List

import numpy as np
from tensorflow.python.keras.utils.data_utils import get_file

from fastestimator.dataset.numpy_dataset import NumpyDataset


def load_data(image_key: str = "x", label_key: str = "y") -> Tuple[NumpyDataset, NumpyDataset]:
    """Load and return the ciFAIR10 dataset.

    This is the cifar10 dataset but with test set duplicates removed and replaced. See
    https://arxiv.org/pdf/1902.00423.pdf or https://cvjena.github.io/cifair/ for details. Cite the paper if you use the
    dataset.

    Args:
        image_key: The key for image.
        label_key: The key for label.

    Returns:
        (train_data, test_data)
    """
    dirname = 'ciFAIR-10'
    archive_name = 'ciFAIR-10.zip'
    origin = 'https://github.com/cvjena/cifair/releases/download/v1.0/ciFAIR-10.zip'
    md5_hash = 'ca08fd390f0839693d3fc45c4e49585f'

    path = get_file(archive_name, origin=origin, file_hash=md5_hash, hash_algorithm='md5', extract=True,
                    archive_format='zip')
    path = os.path.join(os.path.dirname(path), dirname)

    num_train_samples = 50000

    x_train = np.empty((num_train_samples, 3, 32, 32), dtype='uint8')
    y_train = np.empty((num_train_samples,), dtype='uint8')

    for i in range(1, 6):
        fpath = os.path.join(path, f'data_batch_{i}')
        (x_train[(i - 1) * 10000:i * 10000, :, :, :],
         y_train[(i - 1) * 10000:i * 10000]) = _load_batch(fpath)

    fpath = os.path.join(path, 'test_batch')
    x_test, y_test = _load_batch(fpath)

    y_train = np.reshape(y_train, (len(y_train), 1))
    y_test = np.reshape(y_test, (len(y_test), 1))

    x_train = x_train.transpose((0, 2, 3, 1))
    x_test = x_test.transpose((0, 2, 3, 1))

    x_test = x_test.astype(x_train.dtype)
    y_test = y_test.astype(y_train.dtype)

    train_data = NumpyDataset({image_key: x_train, label_key: y_train})
    test_data = NumpyDataset({image_key: x_test, label_key: y_test})
    return train_data, test_data


def _load_batch(file_path: str, label_key: str = 'labels') -> Tuple[np.ndarray, List[int]]:
    """An internal utility for parsing ciFAIR data.

    Args:
        file_path: The path to the file to parse.
        label_key: Key used for the label data within the data dictionary.

    Returns:
        (data, labels)
    """
    with open(file_path, 'rb') as f:
        d = pickle.load(f, encoding='bytes')
        # decode bytes to utf8
        d_decoded = {}
        for k, v in d.items():
            d_decoded[k.decode('utf-8') if isinstance(k, bytes) else k] = v
        d = d_decoded
    data = d['data']
    labels = d[label_key]

    data = data.reshape(data.shape[0], 3, 32, 32)
    return data, labels



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from unittest.mock import patch

import numpy as np

from fastestimator.trace.io import ImageViewer
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData


class TestImageViewer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.image = 0.5 * np.ones(shape=(1, 28, 28, 3))
        mask = np.zeros_like(cls.image)
        mask[0, 10:20, 10:30, :] = [1, 0, 0]
        bbox = np.array([[[3, 7, 10, 6, 'box1'], [20, 20, 8, 8, 'box2']]] * 1)
        cls.img_data = ImgData(y=np.ones((4, )), x=[cls.image, mask, bbox])

    @patch("matplotlib.pyplot.show")
    def test_image_on_epoch_end(self, mock_show):
        data = Data({'x': 0.5 * np.ones(shape=(1, 28, 28, 3))})
        imageviewer = ImageViewer(inputs='x')
        imageviewer.on_epoch_end(data=data)
        self.assertEqual(len(mock_show.mock_calls), 1)

    @patch("matplotlib.pyplot.show")
    def test_imgdata_on_epoch_end(self, mock_show):
        data = Data({'x': self.img_data})
        imageviewer = ImageViewer(inputs='x')
        imageviewer.on_epoch_end(data=data)
        self.assertEqual(len(mock_show.mock_calls), 1)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
import fastestimator.test.unittest_util as fet


class TestToType(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data_np = {
            "x": np.ones((10, 15), dtype="float32"),
            "y": [np.ones((4), dtype="int8"), np.ones((5, 3), dtype="double")],
            "z": {
                "key": np.ones((2, 2), dtype="int64")
            }
        }
        cls.data_tf = {
            "x": tf.ones((10, 15), dtype="float32"),
            "y": [tf.ones((4), dtype="int8"), tf.ones((5, 3), dtype="double")],
            "z": {
                "key": tf.ones((2, 2), dtype="int64")
            }
        }
        cls.data_torch = {
            "x": torch.ones((10, 15), dtype=torch.float32),
            "y": [torch.ones((4), dtype=torch.int8), torch.ones((5, 3), dtype=torch.double)],
            "z": {
                "key": torch.ones((2, 2), dtype=torch.long)
            }
        }
        cls.op_np = {
            'x': np.dtype('float32'), 'y': [np.dtype('int8'), np.dtype('float64')], 'z': {
                'key': np.dtype('int64')
            }
        }
        cls.op_tf = {'x': tf.float32, 'y': [tf.int8, tf.float64], 'z': {'key': tf.int64}}
        cls.op_torch = {'x': torch.float32, 'y': [torch.int8, torch.float64], 'z': {'key': torch.int64}}

    def test_to_type_np(self):
        types = fe.backend.to_type(self.data_np)
        self.assertTrue(fet.is_equal(types, self.op_np))

    def test_to_type_tf(self):
        types = fe.backend.to_type(self.data_tf)
        self.assertTrue(fet.is_equal(types, self.op_tf))

    def test_to_type_torch(self):
        types = fe.backend.to_type(self.data_torch)
        self.assertTrue(fet.is_equal(types, self.op_torch))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.metric import Dice
from fastestimator.util import Data


class TestDice(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        x = np.array([1, 2, 3])
        x_pred = np.array([[1, 1, 3], [2, 3, 4], [1, 1, 0]])
        cls.data = Data({'x': x, 'x_pred': x_pred})
        cls.dice_output = [1.4999999987500001, 2.3999999972, 2.3999999972]
        cls.dice = Dice(true_key='x', pred_key='x_pred')
        cls.dice.system = sample_system_object()

    def test_on_epoch_begin(self):
        self.dice.on_epoch_begin(data=self.data)
        self.assertEqual(self.dice.dice, [])

    def test_on_batch_end(self):
        self.dice.dice = []
        self.dice.on_batch_end(data=self.data)
        self.assertEqual(self.dice.dice, self.dice_output)

    def test_on_epoch_end(self):
        self.dice.dice = [1.4999999987500001, 2.3999999972, 2.3999999972]
        self.dice.on_epoch_end(data=self.data)
        with self.subTest('Check if dice exists'):
            self.assertIn('Dice', self.data)
        with self.subTest('Check the value of dice'):
            self.assertEqual(self.data['Dice'], 2.0999999977166666)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from collections import deque
from copy import deepcopy

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.architecture.pytorch import LeNet as LeNet_torch
from fastestimator.architecture.tensorflow import LeNet as LeNet_tf
from fastestimator.dataset.data import mnist
from fastestimator.op.numpyop.univariate import ExpandDims, Minmax
from fastestimator.op.tensorop.gradient import GradientOp
from fastestimator.op.tensorop.loss import CrossEntropy
from fastestimator.op.tensorop.model import ModelOp, UpdateOp
from fastestimator.schedule import EpochScheduler, RepeatScheduler


class CheckNetworkWeight(fe.trace.Trace):
    """This class will test if the model weight updates correctly.

    Args:
        model: Target model.
        grad_key: Gradients key.
        merge_grad: 'merge_grad' argument of UpdateOp.
        test_self: The object reference of unittest.TestClass
        framework: Which framework the testing target uses. ["tf", "torch"]
        lrs: Learning rate of the optimizer. LR is constant if `lrs` is not iterable. Otherwise the lrs will be
            lrs[x] when work_intervals[x][0] <= epoch < epoch work_intervals[x][1]
        work_intervals: At which interval of epoch does the test activated. This argument need to work with `lrs`.
            When epoch is not in the work_intervals, the test will be skipped. If None, testing will be active all the
            time.
    """
    def __init__(self, model, grad_key, merge_grad, test_self, framework, lrs, work_intervals=None):
        if work_intervals:
            assert len(work_intervals) == len(lrs), "length of work_intervals need to be the same as lrs"
        else:
            assert isinstance(lrs, (int, float)), "if work_intervals is None,  lrs need to be a value"

        super().__init__(inputs=grad_key)
        self.model = model
        self.grad_key = grad_key
        self.merge_grad = merge_grad
        self.test_self = test_self
        self.framework = framework
        self.lrs = lrs
        self.work_intervals = work_intervals
        self.n_gpu = torch.cuda.device_count()
        if self.framework == "tf":
            self.previous_weights = [x.numpy() for x in model.trainable_variables]
        else:
            self.previous_weights = [
                deepcopy(x).cpu().detach().numpy() for x in self.model.parameters() if x.requires_grad
            ]
        self.gradients = deque(maxlen=merge_grad)
        self.new_weight = None

    def on_batch_end(self, data):
        if self.framework == "tf":
            if self.n_gpu > 1:
                # the data[self.key] shape is self.n_gpu times large on axis 0 and need to be folded
                gradients = [self.fold(x.numpy()) for x in data[self.grad_key]]
            else:
                gradients = [x.numpy() for x in data[self.grad_key]]

            self.gradients.append(gradients)
            self.new_weight = [x.numpy() for x in self.model.trainable_variables]
        else:
            self.gradients.append([x.cpu().detach().numpy() for x in data[self.grad_key]])
            self.new_weight = [deepcopy(x).cpu().detach().numpy() for x in self.model.parameters() if x.requires_grad]

        lr = self.get_lr()  # if lr is False, don't need to do the check
        if lr:
            if self.system.global_step % self.merge_grad == 0:
                self.test_self.assertTrue(self.check_if_changed())  # model weight should change
                self.test_self.assertTrue(self.check_if_update_is_gradients(lr))  # model weight should be correct
            else:
                self.test_self.assertFalse(self.check_if_changed())  # model weight should not change

        self.previous_weights = self.new_weight

    def check_if_changed(self):
        """Check if model weight get updated
        """
        for var1, var2 in zip(self.previous_weights, self.new_weight):
            if not np.allclose(var1, var2):
                return True
        return False

    def check_if_update_is_gradients(self, lr):
        """Check if the updated model weight is correct
        """
        for i in range(len(self.previous_weights)):
            diff = self.new_weight[i] - self.previous_weights[i]
            diff2 = sum([x[i] for x in self.gradients]) / self.merge_grad * -lr

            if not np.allclose(diff, diff2, atol=1e-4):
                return False
        return True

    def get_lr(self):
        """Get the current lr. Return False if testing is not active.
        """
        if not self.work_intervals:
            return self.lrs

        for lr, interval in zip(self.lrs, self.work_intervals):
            assert len(interval) == 2
            if interval[0] <= self.system.epoch_idx < interval[1]:
                return lr

        return False

    def fold(self, x):
        """Reduce the size of `x` on axis 0 by self.n_gpu factor.
        """
        assert isinstance(x, np.ndarray)
        assert x.shape[0] % self.n_gpu == 0
        len_org = x.shape[0]
        len_new = int(x.shape[0] / self.n_gpu)
        return sum([x[start:start + len_new] / self.n_gpu for start in range(0, len_org, len_new)])


class TestUpdateOp(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.train_data, _ = mnist.load_data()

    def test_tf_end_to_end(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x"), Minmax(inputs="x", outputs="x")])

            model = fe.build(model_fn=LeNet_tf,
                             optimizer_fn=lambda: tf.optimizers.SGD(lr),
                             mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   lrs=lr,
                                   framework="tf")
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=2,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)

                        else:
                            run_test(mixed_precision, merge_grad, gradient)

    def test_torch_end_to_end(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x", axis=0), Minmax(inputs="x", outputs="x")])

            model = fe.build(model_fn=LeNet_torch,
                             optimizer_fn=lambda x: torch.optim.SGD(params=x, lr=lr),
                             mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   lrs=lr,
                                   framework="torch")
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=2,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)
                        else:
                            run_test(mixed_precision, merge_grad, gradient)

    def test_tf_multi_optimizer_with_epoch_scheduler(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            lr2 = 0.01
            lr3 = 0.001
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x"), Minmax(inputs="x", outputs="x")])

            optimizer_fn = EpochScheduler({
                1: lambda: tf.optimizers.SGD(lr), 2: lambda: tf.optimizers.SGD(lr2), 3: lambda: tf.optimizers.SGD(lr3)
            })

            model = fe.build(model_fn=LeNet_tf, optimizer_fn=optimizer_fn, mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   framework="tf",
                                   lrs=[lr, lr2, lr3],
                                   work_intervals=[[1, 2], [2, 3], [3, 4]])
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=3,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)

                        else:
                            run_test(mixed_precision, merge_grad, gradient)

    def test_torch_multi_optimizer_with_epoch_scheduler(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            lr2 = 0.01
            lr3 = 0.001
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x", axis=0), Minmax(inputs="x", outputs="x")])

            optimizer_fn = EpochScheduler({
                1: lambda x: torch.optim.SGD(params=x, lr=lr),
                2: lambda x: torch.optim.SGD(params=x, lr=lr2),
                3: lambda x: torch.optim.SGD(params=x, lr=lr3)
            })

            model = fe.build(model_fn=LeNet_torch, optimizer_fn=optimizer_fn, mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   framework="torch",
                                   lrs=[lr, lr2, lr3],
                                   work_intervals=[[1, 2], [2, 3], [3, 4]])
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=3,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)

                        else:
                            run_test(mixed_precision, merge_grad, gradient)

    def test_tf_multi_optimizer_with_repeat_scheduler(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            lr2 = 0.01
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x"), Minmax(inputs="x", outputs="x")])

            optimizer_fn = RepeatScheduler([lambda: tf.optimizers.SGD(lr), lambda: tf.optimizers.SGD(lr2)])

            model = fe.build(model_fn=LeNet_tf, optimizer_fn=optimizer_fn, mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   framework="tf",
                                   lrs=[lr, lr2, lr, lr2],
                                   work_intervals=[[1, 2], [2, 3], [3, 4], [4, 5]])
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=4,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)

                        else:
                            run_test(mixed_precision, merge_grad, gradient)

    def test_torch_multi_optimizer_with_repeat_scheduler(self):
        """This test cover the all combination of:
            - mixed-precision / not
            - merge_grad / not
            - gradient input / loss input
        """
        def run_test(mixed_precision, merge_grad, gradient):
            lr = 0.1
            lr2 = 0.01
            pipeline = fe.Pipeline(train_data=self.train_data,
                                   batch_size=4,
                                   ops=[ExpandDims(inputs="x", outputs="x", axis=0), Minmax(inputs="x", outputs="x")])

            optimizer_fn = RepeatScheduler(
                [lambda x: torch.optim.SGD(params=x, lr=lr), lambda x: torch.optim.SGD(params=x, lr=lr2)])

            model = fe.build(model_fn=LeNet_torch, optimizer_fn=optimizer_fn, mixed_precision=mixed_precision)
            network = fe.Network(ops=[
                ModelOp(model=model, inputs="x", outputs="y_pred"),
                CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
                GradientOp(model=model, finals="ce", outputs="grad"),
                UpdateOp(model=model, loss_name="ce", gradients=gradient, merge_grad=merge_grad),
            ])

            traces = [
                CheckNetworkWeight(model=model,
                                   grad_key="grad",
                                   merge_grad=merge_grad,
                                   test_self=self,
                                   framework="torch",
                                   lrs=[lr, lr2, lr, lr2],
                                   work_intervals=[[1, 2], [2, 3], [3, 4], [4, 5]])
            ]
            estimator = fe.Estimator(pipeline=pipeline,
                                     network=network,
                                     epochs=4,
                                     traces=traces,
                                     max_train_steps_per_epoch=2)
            estimator.fit(warmup=False)

        for mixed_precision in [True, False]:
            for merge_grad in [1, 2]:
                for gradient in ["grad", None]:
                    with self.subTest("mixed_precision: {}, merge_grad: {}, take: {}".format(
                            mixed_precision, merge_grad, "gradient" if gradient else "loss")):

                        if (mixed_precision and gradient) or (torch.cuda.device_count() > 1 and merge_grad > 1):
                            with self.assertRaises(ValueError):
                                run_test(mixed_precision, merge_grad, gradient)

                        else:
                            run_test(mixed_precision, merge_grad, gradient)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from unittest.mock import patch

import fastestimator as fe


class TestMain(unittest.TestCase):
    """
    This integration test covers:
    * fe.cli.logs.configure_log_parser
    * fe.cli.train.configure_test_parser
    * fe.cli.train.configure_train_parser
    * fe.cli.run_main
    """
    def test_cli_main_run_train(self):
        with patch('fastestimator.cli.train.train') as fake:
            fe.cli.run_main(["train", "example_entry.py", "--epochs", "3", "--batch_size", "64"])
            args, unknown = fake.call_args[0]

            with self.subTest('args["mode"]'):
                self.assertEqual(args["mode"], "train")

            with self.subTest('args["entry_point"]'):
                self.assertEqual(args["entry_point"], "example_entry.py")

            with self.subTest('args["hyperparameters_json"]'):
                self.assertEqual(args["hyperparameters_json"], None)

            with self.subTest('args["warmup"]'):
                self.assertEqual(args["warmup"], True)

            with self.subTest("unknown"):
                self.assertEqual(unknown, ['--epochs', '3', '--batch_size', '64'])

    def test_cli_main_run_test(self):
        with patch('fastestimator.cli.train.test') as fake:
            fe.cli.run_main(["test", "example_entry.py", "--randon", "200"])
            args, unknown = fake.call_args[0]

            with self.subTest('args["mode"]'):
                self.assertEqual(args["mode"], "test")

            with self.subTest('args["entry_point"]'):
                self.assertEqual(args["entry_point"], "example_entry.py")

            with self.subTest('args["hyperparameters_json"]'):
                self.assertEqual(args["hyperparameters_json"], None)

            with self.subTest("unknown"):
                self.assertEqual(unknown, ["--randon", "200"])

    def test_cli_main_run_logs(self):
        with patch('fastestimator.cli.logs.logs') as fake:
            fe.cli.run_main(["logs", "example_entry.py", "-b", "-v"])
            args, unknown = fake.call_args[0]

            with self.subTest('args["mode"]'):
                self.assertEqual(args["mode"], "logs")

            with self.subTest('args["log_dir"]'):
                self.assertEqual(args["log_dir"], "example_entry.py")

            with self.subTest('args["recursive"]'):
                self.assertEqual(args["recursive"], False)

            with self.subTest('args["ignore"]'):
                self.assertEqual(args["ignore"], None)

            with self.subTest('args["smooth"]'):
                self.assertEqual(args["smooth"], 1)

            with self.subTest('args["pretty_names"]'):
                self.assertEqual(args["pretty_names"], False)

            with self.subTest('args["smooth"]'):
                self.assertEqual(args["smooth"], 1)

            with self.subTest('args["share_legend"]'):
                self.assertEqual(args["share_legend"], True)

            with self.subTest('args["save"]'):
                self.assertEqual(args["save"], False)

            with self.subTest('args["save_dir"]'):
                self.assertEqual(args["save_dir"], None)

            with self.subTest("unknown"):
                self.assertEqual(unknown, ["-b", "-v"])



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Collection, TypeVar, Union

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def to_type(data: Union[Collection, Tensor]) -> Union[Collection, str]:
    """Compute the data types of tensors within a collection of `data` recursively.

    This method can be used with Numpy data:
    ```python
    data = {"x": np.ones((10,15), dtype="float32"), "y":[np.ones((4), dtype="int8"), np.ones((5, 3), dtype="double")],
        "z":{"key":np.ones((2,2), dtype="int64")}}
    types = fe.backend.to_type(data)
    # {'x': dtype('float32'), 'y': [dtype('int8'), dtype('float64')], 'z': {'key': dtype('int64')}}
    ```

    This method can be used with TensorFlow tensors:
    ```python
    data = {"x": tf.ones((10,15), dtype="float32"), "y":[tf.ones((4), dtype="int8"), tf.ones((5, 3), dtype="double")],
        "z":{"key":tf.ones((2,2), dtype="int64")}}
    types = fe.backend.to_type(data)
    # {'x': tf.float32, 'y': [tf.int8, tf.float64], 'z': {'key': tf.int64}}
    ```

    This method can be used with PyTorch tensors:
    ```python
    data = {"x": torch.ones((10,15), dtype=torch.float32), "y":[torch.ones((4), dtype=torch.int8), torch.ones((5, 3),
        dtype=torch.double)], "z":{"key":torch.ones((2,2), dtype=torch.long)}}
    types = fe.backend.to_type(data)
    # {'x': torch.float32, 'y': [torch.int8, torch.float64], 'z': {'key': torch.int64}}
    ```

    Args:
        data: A tensor or possibly nested collection of tensors.

    Returns:
        A collection with the same structure as `data`, but with any tensors substituted for their dtypes.
    """
    if isinstance(data, dict):
        return {key: to_type(value) for (key, value) in data.items()}
    elif isinstance(data, list):
        return [to_type(val) for val in data]
    elif isinstance(data, tuple):
        return tuple([to_type(val) for val in data])
    elif isinstance(data, set):
        return set([to_type(val) for val in data])
    elif hasattr(data, "dtype"):
        return data.dtype
    else:
        return np.array(data).dtype



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, List, TypeVar

import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.backend.roll import roll
from fastestimator.op.tensorop.loss.loss import LossOp

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


class MixLoss(LossOp):
    """Loss class to compute mix-up and cutmix losses.

    This class should be used in conjunction with MixUpBatch and CutMixBatch to perform mix-up training, which helps to
    reduce over-fitting, stabilize GAN training, and harden against adversarial attacks. See
    https://arxiv.org/abs/1710.09412 for details.

    Args:
        loss: A loss object which we use to calculate the underlying loss of MixLoss. This should be an object of type
            fe.op.tensorop.loss.loss.LossOp.
        lam: The key of the lambda value generated by MixUpBatch or CutMixBatch.
        average_loss: Whether the final loss should be averaged or not.

    Raises:
        ValueError: If the provided `loss` has multiple outputs.
    """
    def __init__(self, loss: LossOp, lam: str, average_loss: bool = True):
        self.loss = loss
        self.loss.average_loss = False
        if len(loss.outputs) != 1:
            raise ValueError("MixLoss only supports lossOps which have a single output.")
        super().__init__(inputs=[lam] + loss.inputs,
                         outputs=loss.outputs,
                         mode=loss.mode,
                         ds_id=loss.ds_id,
                         average_loss=average_loss)
        self.out_list = False

    @property
    def pred_key_idx(self) -> int:
        return self.loss.pred_key_idx + 1

    @property
    def true_key_idx(self) -> int:
        return self.loss.true_key_idx + 1

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> Tensor:
        lam, *args = data
        loss1 = self.loss.forward(args, state)

        args[self.loss.true_key_idx] = roll(args[self.loss.true_key_idx], shift=1, axis=0)
        loss2 = self.loss.forward(args, state)

        loss = lam * loss1 + (1.0 - lam) * loss2

        if self.average_loss:
            loss = fe.backend.reduce_mean(loss)

        return loss



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import shutil
from collections import deque
from typing import Optional, Union

import tensorflow as tf
import torch

from fastestimator.backend.save_model import save_model
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable


@traceable()
class ModelSaver(Trace):
    """Save model weights based on epoch frequency during training.

    Args:
        model: A model instance compiled with fe.build.
        save_dir: Folder path into which to save the `model`.
        frequency: Model saving frequency in epoch(s).
        max_to_keep: Maximum number of latest saved files to keep. If 0 or None, all models will be saved.
        save_architecture: Whether to save the full model architecture in addition to the model weights. This option is
            only available for TensorFlow models at present, and will generate a folder containing several files. The
            model can then be re-instantiated even without access to the original code by calling:
            tf.keras.models.load_model(<path to model folder>).

    Raises:
        ValueError: If `max_to_keep` is negative, or if save_architecture is used with a PyTorch model.
    """
    def __init__(self,
                 model: Union[tf.keras.Model, torch.nn.Module],
                 save_dir: str,
                 frequency: int = 1,
                 max_to_keep: Optional[int] = None,
                 save_architecture: bool = False) -> None:
        super().__init__(mode="train")
        self.model = model
        self.save_dir = save_dir
        self.frequency = frequency
        self.save_architecture = save_architecture
        if save_architecture and isinstance(model, torch.nn.Module):
            raise ValueError("Sorry, architecture saving is not currently enabled for PyTorch")
        if max_to_keep is not None and max_to_keep < 0:
            raise ValueError(f"max_to_keep should be a non-negative integer, but got {max_to_keep}")
        self.file_queue = deque([None] * (max_to_keep or 0), maxlen=max_to_keep or 0)

    def on_epoch_end(self, data: Data) -> None:
        # No model will be saved when save_dir is None, which makes smoke test easier.
        if self.save_dir and self.system.epoch_idx % self.frequency == 0:
            model_name = "{}_epoch_{}".format(self.model.model_name, self.system.epoch_idx)
            model_path = save_model(model=self.model,
                                    save_dir=self.save_dir,
                                    model_name=model_name,
                                    save_architecture=self.save_architecture)
            print("FastEstimator-ModelSaver: Saved model to {}".format(model_path))
            rm_path = self.file_queue[self.file_queue.maxlen - 1] if self.file_queue.maxlen else None
            if rm_path:
                os.remove(rm_path)
                if self.save_architecture:
                    shutil.rmtree(os.path.splitext(rm_path)[0])
                print("FastEstimator-ModelSaver: Removed model {} due to file number exceeding max_to_keep".format(
                    rm_path))
            self.file_queue.appendleft(model_path)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import time
import unittest
from io import StringIO
from unittest.mock import patch

import numpy as np
import tensorflow as tf
import torch
from tensorflow.python.client import device_lib

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestParseStringToPython(unittest.TestCase):
    def test_parse_string_to_python_positive_int(self):
        x = fe.util.parse_string_to_python("5")
        self.assertEqual(x, 5)

    def test_parse_string_to_python_negative_float(self):
        x = fe.util.parse_string_to_python("-1.5")
        self.assertEqual(x, -1.5)

    def test_parse_string_to_python_list(self):
        x = fe.util.parse_string_to_python("[5, 4, 0.3]")
        self.assertTrue(x, [5, 4, 0.3])

    def test_parse_string_to_python_list_over_mixed_entries(self):
        x = fe.util.parse_string_to_python("[5, [1,2,3]]")
        self.assertTrue(x, [5, [1, 2, 3]])

    def test_parse_string_to_python_dict(self):
        x = fe.util.parse_string_to_python("{'a': 5, 'b': 7}")
        self.assertEqual(x, {'a': 5, 'b': 7})

    def test_parse_string_to_python_dict_over_mixed_entries(self):
        x = fe.util.parse_string_to_python("{'a':5, 'b':[1,2,3]}")
        self.assertEqual(x, {'a': 5, 'b': [1, 2, 3]})


class TestToList(unittest.TestCase):
    def test_to_list_input_none(self):
        x = fe.util.to_list(None)
        self.assertEqual(x, [])

    def test_to_list_input_float(self):
        x = fe.util.to_list(0.5)
        self.assertEqual(x, [0.5])

    def test_to_list_input_string(self):
        x = fe.util.to_list("a")
        self.assertEqual(x, ["a"])

    def test_to_list_input_list(self):
        x = fe.util.to_list([None, "2", (3, 4), {5, 6}, {"7": 7, "8": 8}, [9]])
        self.assertEqual(x, [None, "2", (3, 4), {5, 6}, {"7": 7, "8": 8}, [9]])

    def test_to_list_input_tuple(self):
        x = fe.util.to_list((None, "2", (3, 4), {5, 6}, {"7": 7, "8": 8}, [9]))
        self.assertEqual(x, [None, "2", (3, 4), {5, 6}, {"7": 7, "8": 8}, [9]])

    def test_to_list_input_set(self):
        x = fe.util.to_list({3, 4})
        self.assertEqual(x, [3, 4])

    def test_to_list_input_dict(self):
        x = fe.util.to_list({"ans": 1})
        self.assertEqual(x, [{"ans": 1}])


class TestToSet(unittest.TestCase):
    def test_to_set_input_none(self):
        x = fe.util.to_set(None)
        self.assertEqual(x, set())

    def test_to_set_input_non_iterable(self):
        x = fe.util.to_set(0)
        self.assertEqual(x, {0})

        x = fe.util.to_set("a")
        self.assertEqual(x, {"a"})

    def test_to_set_input_list(self):
        x = fe.util.to_set([None, "2", (3, 4)])
        self.assertEqual(x, {None, "2", (3, 4)})

    def test_to_set_input_tuple(self):
        x = fe.util.to_set((None, "2", (3, 4)))
        self.assertEqual(x, {None, "2", (3, 4)})

    def test_to_set_input_set(self):
        x = fe.util.to_set({3, 4})
        self.assertEqual(x, {3, 4})


class TestParamToRange(unittest.TestCase):
    def test_param_to_range_int(self):
        x = fe.util.param_to_range(3)
        self.assertEqual(x, (-3, 3))

    def test_param_to_range_float(self):
        x = fe.util.param_to_range(2.3)
        self.assertEqual(x, (-2.3, 2.3))

    def test_param_to_range_neg_int(self):
        x = fe.util.param_to_range(-3)
        self.assertEqual(x, (-3, 3))

    def test_param_to_range_neg_float(self):
        x = fe.util.param_to_range(-2.3)
        self.assertEqual(x, (-2.3, 2.3))

    def test_param_to_range_int_tuple(self):
        x = fe.util.param_to_range((2, 3))
        self.assertEqual(x, (2, 3))

    def test_param_to_range_float_tuple(self):
        x = fe.util.param_to_range((1.2, 3.3))
        self.assertEqual(x, (1.2, 3.3))


class TestNonContext(unittest.TestCase):
    def test_non_context_syntax_work(self):
        a = 5
        with fe.util.NonContext():
            a = a + 37
            self.assertEqual(a, 42)


class TestSuppressor(unittest.TestCase):
    def test_suppressor(self):
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            with fe.util.Suppressor():
                print("hello world")
            log = fake_stdout.getvalue()
            self.assertEqual(log, '')


class TestTimer(unittest.TestCase):
    def test_timer_as_context_manager(self):
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            with patch('tensorflow.print', new=print):
                with fe.util.Timer():
                    time.sleep(1)
                context = fake_stdout.getvalue()  # Tasks took ? seconds
        exec_time = float(context.split(" ")[2])
        self.assertTrue(abs(exec_time - 1) < 0.1)

    def test_timer_as_decorator(self):
        @fe.util.Timer("T2")
        def func():
            time.sleep(1)

        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            with patch('tensorflow.print', new=print):
                func()
                context = fake_stdout.getvalue()  # T2 took ? seconds

        with self.subTest("test the printed name"):
            task_name = context.split(" ")[0]
            self.assertEqual(task_name, "T2")

        with self.subTest("test recorded time"):
            exec_time = float(context.split(" ")[2])
            self.assertTrue(abs(exec_time - 1) < 0.1)


class TestDraw(unittest.TestCase):
    def test_draw_stdout(self):
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            fe.util.draw()
            log = fake_stdout.getvalue()

        expected = ('    ______           __  ______     __  _                 __            \n'
                    '   / ____/___ ______/ /_/ ____/____/ /_(_)___ ___  ____ _/ /_____  _____\n'
                    '  / /_  / __ `/ ___/ __/ __/ / ___/ __/ / __ `__ \\/ __ `/ __/ __ \\/ ___/\n'
                    ' / __/ / /_/ (__  ) /_/ /___(__  ) /_/ / / / / / / /_/ / /_/ /_/ / /    \n'
                    '/_/    \\__,_/____/\\__/_____/____/\\__/_/_/ /_/ /_/\\__,_/\\__/\\____/_/     \n'
                    '                                                                        \n'
                    '\n')

        self.assertEqual(log, expected)


class TestPrettifyMetricName(unittest.TestCase):
    def test_prettify_metric_name(self):
        x = fe.util.prettify_metric_name("myUgly_loss")
        self.assertEqual(x, "My Ugly Loss")

    def test_prettify_metric_name_all_capital(self):
        x = fe.util.prettify_metric_name("ABCD")
        self.assertEqual(x, "Abcd")

    def test_prettify_metric_name_two_space(self):
        x = fe.util.prettify_metric_name("hello  world")
        self.assertEqual(x, "Hello World")


class TestStripSuffix(unittest.TestCase):
    def test_strip_suffix_match(self):
        x = fe.util.strip_suffix("astring.json", ".json")
        self.assertEqual(x, "astring")

    def test_strip_suffix_not_match(self):
        x = fe.util.strip_suffix("astring.json", ".yson")
        self.assertEqual(x, "astring.json")


class TestStripPrefix(unittest.TestCase):
    def test_strip_prefix_match(self):
        x = fe.util.strip_prefix("astring.json", "ast")  # "ring.json"
        self.assertEqual(x, "ring.json")

    def test_strip_prefix_not_match(self):
        x = fe.util.strip_prefix("astring.json", "asa")
        self.assertEqual(x, "astring.json")


class TestGetType(unittest.TestCase):
    def test_get_type_np(self):
        x = fe.util.get_type(np.ones((10, 10), dtype='int32'))
        self.assertEqual(x, 'int32')

        x = fe.util.get_type(np.ones((10, 10), dtype=np.float32))
        self.assertEqual(x, 'float32')

    def test_get_type_tf(self):
        x = fe.util.get_type(tf.ones((10, 10), dtype='float16'))
        self.assertEqual(x, "<dtype: 'float16'>")

        x = fe.util.get_type(tf.Variable([1, 2, 3], dtype=tf.int64))
        self.assertEqual(x, "<dtype: 'int64'>")

    def test_get_type_torch(self):
        x = fe.util.get_type(torch.ones((10, 10)).type(torch.float))
        self.assertEqual(x, "torch.float32")

        x = fe.util.get_type(torch.tensor((10, 10), dtype=torch.int16))
        self.assertEqual(x, "torch.int16")

    def test_get_type_list(self):
        x = fe.util.get_type([np.ones((10, 10)) for i in range(4)])
        self.assertEqual(x, "List[float64]")

        x = fe.util.get_type([1, "a"])  # "List[int]"
        self.assertEqual(x, "List[int]")

        x = fe.util.get_type([[[1]]])  # "List[List[List[int]]]"
        self.assertEqual(x, "List[List[List[int]]]")

    def test_get_type_tuple(self):
        x = fe.util.get_type((tf.ones((10, 10), dtype='float16'), ))
        self.assertEqual(x, "List[<dtype: 'float16'>]")


class TestGetShape(unittest.TestCase):
    def test_get_shape_np_dimension_match(self):
        x = fe.util.get_shape(np.ones((12, 22, 11)))
        self.assertEqual(x, [12, 22, 11])

        x = fe.util.get_shape([np.ones((12, 22, 11))])
        self.assertEqual(x, [None, 12, 22, 11])

        x = fe.util.get_shape([np.ones((12, 22, 11)), np.ones((12, 22, 11))])
        self.assertEqual(x, [None, 12, 22, 11])

        x = fe.util.get_shape([np.ones((12, 22, 11)), np.ones((12, 22, 4))])
        self.assertEqual(x, [None, 12, 22, None])

        x = fe.util.get_shape([np.ones((12, 22, 11)), np.ones((18, 5, 4))])
        self.assertEqual(x, [None, None, None, None])

    def test_get_shape_np_dimension_mismatch(self):
        x = fe.util.get_shape([np.ones((12, 22, 11)), np.ones((18, 5))])
        self.assertEqual(x, [None])


class TestParseModes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.modes = {"train", "eval", "test", "infer"}
        cls.n_modes = {"!train", "!eval", "!test", "!infer"}
        cls.n_modes_map = {
            "!train": {"eval", "test", "infer"},
            "!eval": {"train", "test", "infer"},
            "!test": {"train", "eval", "infer"},
            "!infer": {"train", "eval", "test"}
        }

    def test_parse_modes_single_mode_direct(self):
        for mode in self.modes:
            with self.subTest(modes={mode}):
                self.assertEqual(fe.util.parse_modes({mode}), {mode})

    def test_parse_modes_single_mode_negation(self):
        for mode in self.n_modes:
            with self.subTest(modes={mode}):
                self.assertEqual(fe.util.parse_modes({mode}), self.n_modes_map[mode])

    def test_parse_modes_mutilple_mode_direct(self):
        modes = [{"train", "eval"}, {"train", "eval", "test"}, {"train", "eval", "test", "infer"}]
        for mode in modes:
            with self.subTest(modes=mode):
                self.assertEqual(fe.util.parse_modes(mode), mode)

    def test_parse_modes_mutilple_mode_negation(self):
        modes = [{"!train", "!eval"}, {"!train", "!eval", "!test"}, {"!train", "!eval", "!test", "!infer"}]
        anss = [{"test", "infer"}, {"infer"}, set()]
        for mode, ans in zip(modes, anss):
            with self.subTest(modes=mode):
                self.assertEqual(fe.util.parse_modes(mode), ans)

    def test_parse_modes_invalid_mode(self):
        with self.assertRaises(AssertionError):
            fe.util.parse_modes({"wrong_mode"})

    def test_parse_modes_empty_set(self):
        self.assertEqual(fe.util.parse_modes(set()), set())

    def test_parse_modes_mix_true_and_negation(self):
        with self.assertRaises(AssertionError):
            fe.util.parse_modes({"train", "!eval"})


class TestPadBatch(unittest.TestCase):
    def test_pad_batch_pad_one_entry(self):
        data = [{"x": np.ones((2, 2)), "y": 8}, {"x": np.ones((3, 1)), "y": 4}]
        fe.util.pad_batch(data, pad_value=0)
        obj = [{
            "x": np.array([[1., 1.], [1., 1.], [0., 0.]]), "y": 8
        }, {
            "x": np.array([[1., 0.], [1., 0.], [1., 0.]]), "y": 4
        }]

        self.assertTrue(is_equal(data, obj))

    def test_pad_batch_pad_all_entry(self):
        data = [{"x": np.ones((3, 1)), "y": np.ones((1, 1))}, {"x": np.ones((2, 2)), "y": np.ones((1, 3))}]
        fe.util.pad_batch(data, pad_value=0)
        obj = [{
            "x": np.array([[1., 0.], [1., 0.], [1., 0.]]), "y": np.array([[1., 0., 0.]])
        }, {
            "x": np.array([[1., 1.], [1., 1.], [0., 0.]]), "y": np.array([[1., 1., 1.]])
        }]

        self.assertTrue(is_equal(data, obj))

    def test_pad_batch_different_different_key_assertion(self):
        data = [{"x1": np.ones((2, 2)), "y": 8}, {"x": np.ones((3, 1)), "y": 4}]
        with self.assertRaises(AssertionError):
            fe.util.pad_batch(data, pad_value=0)

    def test_pad_batch_different_rank_mismatch_assertion(self):
        data = [{"x1": np.ones((2, 2, 2)), "y": 8}, {"x": np.ones((3, 1)), "y": 4}]
        with self.assertRaises(AssertionError):
            fe.util.pad_batch(data, pad_value=0)


class TestPadData(unittest.TestCase):
    def test_pad_data_target_shape_all_dimension_larger(self):
        x = np.ones((1, 2))
        x = fe.util.pad_data(x, target_shape=(3, 3), pad_value=-2)
        y = np.array([[1, 1, -2], [-2, -2, -2], [-2, -2, -2]])
        self.assertTrue(is_equal(x, y))

    def test_pad_data_target_shape_has_higher_rank(self):
        x = np.ones((1, 2))
        with self.assertRaises(ValueError):
            x = fe.util.pad_data(x, target_shape=(3, 3, 3), pad_value=-2)

    def test_pad_data_target_shape_has_smaller_rank(self):
        x = np.ones((1, 2))
        with self.assertRaises(ValueError):
            x = fe.util.pad_data(x, target_shape=(1, ), pad_value=-2)

    def test_pad_data_target_shape_all_dimension_smaller(self):
        x = np.ones((3, 3))
        with self.assertRaises(ValueError):
            x = fe.util.pad_data(x, target_shape=(1, 2), pad_value=-2)

    def test_pad_data_target_shape_some_dimension_smaller(self):
        x = np.ones((3, 3))
        with self.assertRaises(ValueError):
            x = fe.util.pad_data(x, target_shape=(1, 4), pad_value=-2)


class TestIsNumber(unittest.TestCase):
    def test_is_number_pos_float(self):
        x = fe.util.is_number("13.7")
        self.assertTrue(x)

    def test_is_number_neg_float(self):
        x = fe.util.is_number("-8.64")
        self.assertTrue(x)

    def test_is_number_pos_scientific_expression(self):
        x = fe.util.is_number("2.5e-10")
        self.assertTrue(x)

    def test_is_number_neg_scientific_expression(self):
        x = fe.util.is_number("-2.5e5")
        self.assertTrue(x)

    def test_is_number_string(self):
        x = fe.util.is_number("apple")
        self.assertFalse(x)

    def test_is_number_string_with_number(self):
        x = fe.util.is_number("123hello")
        self.assertFalse(x)

    def test_is_number_empty_string(self):
        x = fe.util.is_number("")
        self.assertFalse(x)


class TestDefaultKeyDict(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_dict = fe.util.DefaultKeyDict(default=lambda x: x + x, a=4, b=6)

    def test_default_key_dict_use_existed_key(self):
        self.assertEqual(self.test_dict["a"], 4)

    def test_default_key_dict_use_non_existed_key(self):
        self.assertEqual(self.test_dict[10], 20)

    def test_default_key_dict_write_value_on_existed_key(self):
        self.test_dict["a"] = -100
        self.assertEqual(self.test_dict["a"], -100)

    def test_default_key_dict_write_value_on_non_existed_key(self):
        self.test_dict["c"] = "hello"
        self.assertEqual(self.test_dict["c"], "hello")


class TestGetNumDevices(unittest.TestCase):
    def test_get_num_devices(self):
        x = fe.util.get_num_devices()
        local_device_protos = device_lib.list_local_devices()
        gpu_list = [x.name for x in local_device_protos if x.device_type == 'GPU']
        ans = max(1, len(gpu_list))
        self.assertEqual(x, ans)


class TestGetBatchSize(unittest.TestCase):
    def test_get_batch_size_np(self):
        data = {"a": np.ones([3, 4, 5])}
        batch_size = fe.util.get_batch_size(data)
        self.assertEqual(batch_size, 3)

    def test_get_batch_size_tf(self):
        data = {"a": tf.ones([3, 4, 5])}
        batch_size = fe.util.get_batch_size(data)
        self.assertEqual(batch_size, 3)

    def test_get_batch_size_torch(self):
        data = {"a": torch.ones([3, 4, 5])}
        batch_size = fe.util.get_batch_size(data)
        self.assertEqual(batch_size, 3)

    def test_get_batch_size_np_different_shape(self):
        data = {"a": np.ones([3, 4, 5]), "b": np.ones([1, 2])}
        with self.assertRaises(AssertionError):
            batch_size = fe.util.get_batch_size(data)

    def test_get_batch_size_list(self):
        data = [np.ones([3, 4, 5])]
        with self.assertRaises(AssertionError):
            batch_size = fe.util.get_batch_size(data)

    def test_get_batch_size_dict_with_all_irrelevent_value(self):
        data = {"a": 1, "b": "hello"}
        with self.assertRaises(AssertionError):
            batch_size = fe.util.get_batch_size(data)


class TestToNumber(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.n = np.array([1, 2, 3])
        cls.t = tf.constant([1, 2, 3])
        cls.p = torch.tensor([1, 2, 3])

    def test_to_number_np_value(self):
        self.assertTrue(np.allclose(fe.util.util.to_number(self.n), self.n))

    def test_to_number_np_type(self):
        self.assertEqual(type(fe.util.util.to_number(self.n)), np.ndarray)

    def test_to_number_tf_value(self):
        self.assertTrue(np.allclose(fe.util.util.to_number(self.t), self.n))

    def test_to_number_tf_type(self):
        self.assertEqual(type(fe.util.util.to_number(self.t)), np.ndarray)

    def test_to_number_torch_value(self):
        self.assertTrue(np.allclose(fe.util.util.to_number(self.p), self.n))

    def test_to_number_torch_type(self):
        self.assertEqual(type(fe.util.util.to_number(self.p)), np.ndarray)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split

from fastestimator.dataset.numpy_dataset import NumpyDataset


def load_data() -> Tuple[NumpyDataset, NumpyDataset]:
    """Load and return the UCI ML Breast Cancer Wisconsin (Diagnostic) dataset.

    For more information about this dataset and the meaning of the features it contains, see the sklearn documentation.

    Returns:
        (train_data, eval_data)
    """
    (x, y) = load_breast_cancer(return_X_y=True)
    x_train, x_eval, y_train, y_eval = train_test_split(x, y, test_size=0.2, random_state=42)
    x_train, x_eval = np.float32(x_train), np.float32(x_eval)
    train_data = NumpyDataset({"x": x_train, "y": y_train})
    eval_data = NumpyDataset({"x": x_eval, "y": y_eval})
    return train_data, eval_data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestPercentile(unittest.TestCase):
    def test_percentile_tf_input_axis_none(self):
        with self.subTest("even_elements"):
            t = tf.constant([1, 2])
            obj1 = fe.backend.percentile(t, percentiles=50)
            obj2 = tf.constant([1])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("two_dimensional"):
            t = tf.constant([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50)
            obj2 = tf.constant([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50])
            obj2 = tf.constant([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_tf_input_axis_not_none(self):
        with self.subTest("two_dimensional"):
            t = tf.constant([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50, axis=0)
            obj2 = tf.constant([[2, 4, 6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=1)
            obj2 = tf.constant([[3], [5], [6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=[0, 1])
            obj2 = tf.constant([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50], axis=[0, 1])
            obj2 = tf.constant([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_tf_input_axis_not_none_keepdims_false(self):
        with self.subTest("two_dimensional"):
            t = tf.constant([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50, axis=0, keepdims=False)
            obj2 = tf.constant([2, 4, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=1, keepdims=False)
            obj2 = tf.constant([3, 5, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=[0, 1], keepdims=False)
            obj2 = tf.constant(5)
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50], axis=[0, 1], keepdims=False)
            obj2 = tf.constant([1, 5])
            self.assertTrue(is_equal(obj1, obj2))

    # ------------------------- torch input --------------------------------------
    def test_percentile_torch_input_axis_none(self):
        with self.subTest("even_elements"):
            t = torch.tensor([1, 2])
            obj1 = fe.backend.percentile(t, percentiles=50)
            obj2 = torch.tensor([1])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("two_dimensional"):
            t = torch.tensor([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50)
            obj2 = torch.tensor([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50])
            obj2 = torch.tensor([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_torch_input_axis_not_none(self):
        with self.subTest("two_dimensional"):
            t = torch.tensor([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50, axis=0)
            obj2 = torch.tensor([[2, 4, 6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=1)
            obj2 = torch.tensor([[3], [5], [6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=[0, 1])
            obj2 = torch.tensor([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50], axis=[0, 1])
            obj2 = torch.tensor([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_torch_input_axis_not_none_keepdims_false(self):
        with self.subTest("two_dimensional"):
            t = torch.tensor([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(t, percentiles=50, axis=0, keepdims=False)
            obj2 = torch.tensor([2, 4, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=1, keepdims=False)
            obj2 = torch.tensor([3, 5, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(t, percentiles=50, axis=[0, 1], keepdims=False)
            obj2 = torch.tensor(5)
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(t, percentiles=[0, 50], axis=[0, 1], keepdims=False)
            obj2 = torch.tensor([1, 5])
            self.assertTrue(is_equal(obj1, obj2))

    # ------------------------- numpy input ---------------------------------------
    def test_percentile_np_input_axis_none(self):
        with self.subTest("even_elements"):
            n = np.array([1, 2])
            obj1 = fe.backend.percentile(n, percentiles=50)
            obj2 = np.array([1])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("two_dimensional"):
            n = np.array([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(n, percentiles=50)
            obj2 = np.array([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(n, percentiles=[0, 50])
            obj2 = np.array([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_np_input_axis_not_none(self):
        with self.subTest("two_dimensional"):
            n = np.array([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(n, percentiles=50, axis=0)
            obj2 = np.array([[2, 4, 6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(n, percentiles=50, axis=1)
            obj2 = np.array([[3], [5], [6]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(n, percentiles=50, axis=[0, 1])
            obj2 = np.array([[5]])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(n, percentiles=[0, 50], axis=[0, 1])
            obj2 = np.array([[[1]], [[5]]])
            self.assertTrue(is_equal(obj1, obj2))

    def test_percentile_np_input_axis_not_none_keepdims_false(self):
        with self.subTest("two_dimensional"):
            n = np.array([[1, 3, 9], [2, 7, 5], [8, 4, 6]])
            obj1 = fe.backend.percentile(n, percentiles=50, axis=0, keepdims=False)
            obj2 = np.array([2, 4, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("single_axis"):
            obj1 = fe.backend.percentile(n, percentiles=50, axis=1, keepdims=False)
            obj2 = np.array([3, 5, 6])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("multi_axis"):
            obj1 = fe.backend.percentile(n, percentiles=50, axis=[0, 1], keepdims=False)
            self.assertTrue(is_equal(obj1, 5, assert_type=False))

        with self.subTest("multi_percentile"):
            obj1 = fe.backend.percentile(n, percentiles=[0, 50], axis=[0, 1], keepdims=False)
            obj2 = np.array([1, 5])
            self.assertTrue(is_equal(obj1, obj2))
